# SPDX-FileCopyrightText: (c) 2024 The Tenzir Contributors
# SPDX-License-Identifier: BSD-3-Clause

"""Usage:
  tenzir-platform node list [--json]
  tenzir-platform node ping <node>
  tenzir-platform node create [--name=<node_name>]
  tenzir-platform node config <node> [-o filename] [--format=docker|tenzir|tenzir-node]
  tenzir-platform node delete <node>
  tenzir-platform node run [--name=<node_name>] [--image=<container_image>]

Options:
  --name=<node_name>         The name of the newly created node [default: CLI_Node]
  --image=<container_image>  The docker image to use for the newly created node
  -o,--output=<filename>     Where to write the config file. Set to "-" to write to stdout.
  --format=<config_type>       The format of the downloaded config file. [default: docker]

Description:
  tenzir-platform node list [--json]
    Display a list of all nodes in the current workspace.

  tenzir-platform node ping
    Send a ping request to a node and measure the response time.

  tenzir-platform node create
    Create a new node.

  tenzir-platform node config <node> [-o filename] [--format=docker|tenzir|tenzir-node]
    Download the configuration for an existing node.
    The special filename '-' can be used to print to stdout instead of a file.

  tenzir-platform node delete <node>
    Delete the specified node.

  tenzir-platform node run
    Create a new temporary node and run it in a local 'docker compose' stack.
    A new node is created in the currently selected workspace.
    The node is deleted when the command is interrupted.
    Requires the 'docker compose' binary in the current PATH.
"""

from tenzir_platform.helpers.cache import load_current_workspace
from tenzir_platform.helpers.client import AppClient
from tenzir_platform.helpers.environment import PlatformEnvironment
from pydantic import BaseModel
from docopt import docopt
from typing import Optional, List
from requests import HTTPError
import json
import time
import tempfile
import os
import subprocess
import re
import random
import datetime


def _is_node_id(identifier: str):
    return bool(re.match(r"^n-[a-z0-9]{8}$", identifier))


def _get_node_list(client: AppClient, workspace_id: str) -> List:
    resp = client.post(
        "list-nodes",
        json={
            "tenant_id": workspace_id,
        },
    )
    resp.raise_for_status()
    return resp.json()["nodes"]


def _resolve_node_identifier(
    client: AppClient, workspace_id: str, identifier: str
) -> str:
    # If we already have a node id, use that.
    if _is_node_id(identifier):
        return identifier

    # Otherwise go through the list of nodes and look for a matching name.
    nodes = _get_node_list(client, workspace_id)
    name_matched = [node for node in nodes if node["name"] == identifier]
    if len(name_matched) == 0:
        raise Exception(f"Unknown node {identifier}")
    if len(name_matched) > 1:
        matching_ids = [node["node_id"] for node in name_matched]
        raise Exception(f"Ambigous name {identifier} is shared by nodes {matching_ids}")
    return name_matched[0]["node_id"]


def list(client: AppClient, workspace_id: str, print_json: bool):
    nodes = _get_node_list(client, workspace_id)
    if print_json:
        print(f"{json.dumps(nodes)}")
    else:
        for i, node in enumerate(nodes):
            connection_symbol = "🟢" if node["lifecycle_state"] == "connected" else "🔴"
            print(f"{connection_symbol} {node['name']} ({node['node_id']})")


def ping(client: AppClient, workspace_id: str, node: str):
    node_id = _resolve_node_identifier(client, workspace_id, node)
    start = time.time()
    resp = client.post(
        "proxy",
        json={
            "node_id": node_id,
            "tenant_id": workspace_id,
            "http": {
                "path": "/ping",
                "method": "POST",
            },
        },
    )
    if resp.status_code == 410:
        print("node disconnected")
        return
    elif resp.status_code != 200:
        print(f"error {resp.status_code} after {time.time()-start}s")
        return
    print(f"response {resp.status_code} in {time.time()-start}s")


def run(
    client: AppClient, workspace_id: str, node_name: str, container_image: Optional[str]
):
    resp = client.post(
        "generate-client-config",
        json={
            "config_type": "docker",
            "tenant_id": workspace_id,
            "node_name": node_name,
        },
    )
    resp.raise_for_status()
    json = resp.json()
    node_id = json["node_id"]
    contents = json["contents"]

    # Hacky text replacement for using custom containers
    if container_image is not None:
        pattern = r"^(\s*image:) tenzir/tenzir.*$"
        replacement = r"\1 " + container_image
        contents = re.sub(pattern, replacement, contents, count=1, flags=re.MULTILINE)

    temp_file = tempfile.NamedTemporaryFile(delete=False, mode="w+")
    temp_file_name = temp_file.name
    temp_file.write(contents)
    temp_file.flush()

    try:
        print("running temporary Tenzir node")
        full_command = f"docker compose -f {temp_file_name} up"
        subprocess.run(full_command, shell=True, check=True)
    except KeyboardInterrupt:
        # Regular exit with CTRL-C
        pass
    except subprocess.CalledProcessError as e:
        print(f"Error running the docker compose command: {e}")
    finally:
        print("removing node and config file")
        if os.path.exists(temp_file_name):
            os.remove(temp_file_name)
        client.post(
            "delete-node",
            json={
                "tenant_id": workspace_id,
                "node_id": node_id,
            },
        )


def create(client: AppClient, workspace_id: str, node_name: Optional[str]):
    if node_name is None:
        time_suffix = datetime.datetime.now(datetime.timezone.utc).strftime(
            "%Y%m%dT%H%M%SZ"
        )
        node_name = f"node-{time_suffix}"
    resp = client.post(
        "generate-client-config",
        json={
            "tenant_id": workspace_id,
            "config_type": "docker",
            "node_name": node_name,
        },
    )
    resp.raise_for_status()
    print(f"created node {resp.json()['node_id']}")


def config(
    client: AppClient,
    workspace_id: str,
    node: str,
    config_format: str,
    output_file: Optional[str],
):
    node_id = _resolve_node_identifier(client, workspace_id, node)
    resp = client.post(
        "generate-client-config",
        json={
            "tenant_id": workspace_id,
            "config_type": config_format,
            "node_id": node_id,
        },
    )
    resp.raise_for_status()
    response_body = resp.json()
    node_id = response_body["node_id"]
    if output_file is None:
        output_file = response_body["filename"]
    contents = response_body["contents"]
    if output_file == "-":
        print(contents)
    else:
        with open(output_file, "w") as f:
            f.write(contents)
        print(f"wrote config for node {node_id} to file {output_file}")


def delete(client: AppClient, workspace_id: str, node: str):
    node_id = _resolve_node_identifier(client, workspace_id, node)
    resp = client.post(
        "delete-node",
        json={
            "tenant_id": workspace_id,
            "node_id": node_id,
        },
    )
    resp.raise_for_status()
    print(f"deleted node {node_id}")


def node_subcommand(platform: PlatformEnvironment, argv):
    args = docopt(__doc__, argv=argv)
    try:
        workspace_id, user_key = load_current_workspace(platform)
        client = AppClient(platform=platform)
        client.workspace_login(user_key)
    except Exception as e:
        print(f"error: {e}")
        print(
            "Failed to load current workspace, please run 'tenzir-platform workspace select' first"
        )
        exit(1)

    try:
        if args["list"]:
            json = args["--json"]
            list(client, workspace_id, json)
        elif args["ping"]:
            node = args["<node>"]
            ping(client, workspace_id, node)
        elif args["run"]:
            node_name = args["--name"]
            container_image = args["--image"]
            run(client, workspace_id, node_name, container_image)
        elif args["create"]:
            node_name = args["--name"]
            create(client, workspace_id, node_name)
        elif args["config"]:
            node = args["<node>"]
            format_ = args["--format"]
            output = args["--output"]
            config(client, workspace_id, node, format_, output)
        elif args["delete"]:
            node = args["<node>"]
            delete(client, workspace_id, node)
    except HTTPError as e:
        if e.response.status_code == 403:
            print(
                "Access denied. Please try re-authenticating by running 'tenzir-platform workspace select'"
            )



# SPDX-FileCopyrightText: (c) 2024 The Tenzir Contributors
# SPDX-License-Identifier: BSD-3-Clause

"""Usage:
  tenzir-platform alert add <node> <duration> <webhook_url> [<webhook_body>]
  tenzir-platform alert delete <alert_id>
  tenzir-platform alert list

Options:
  <node>         The node to be monitored.
  <duration>     The amount of time to wait before triggering the alert.
  <webhook_url>  The URL to call when the alert triggers
  <webhook_body> The body to send along with the webhook. Must be valid JSON.

Description:
  tenzir-platform alert add <node> <duration> <webhook>
    Add a new alert to the platform.

  tenzir-platform alert delete <alert_id>
    Delete the specified alert.

  tenzir-platform alert list
    List all configured alerts.
"""

from tenzir_platform.helpers.cache import load_current_workspace
from tenzir_platform.helpers.client import AppClient
from tenzir_platform.helpers.environment import PlatformEnvironment
from pydantic import BaseModel
from docopt import docopt
from typing import Optional, List
from requests import HTTPError
from pytimeparse2 import parse as parse_duration
import json
import time
import tempfile
import os
import subprocess
import re
import random
import datetime


def _is_node_id(identifier: str):
    return bool(re.match(r"^n-[a-z0-9]{8}$", identifier))


def _get_node_list(client: AppClient, workspace_id: str) -> List:
    resp = client.post(
        "list-nodes",
        json={
            "tenant_id": workspace_id,
        },
    )
    resp.raise_for_status()
    return resp.json()["nodes"]


def _resolve_node_identifier(
    client: AppClient, workspace_id: str, identifier: str
) -> str:
    # If we already have a node id, use that.
    if _is_node_id(identifier):
        return identifier

    # Otherwise go through the list of nodes and look for a matching name.
    nodes = _get_node_list(client, workspace_id)
    name_matched = [node for node in nodes if node["name"] == identifier]
    if len(name_matched) == 0:
        raise Exception(f"Unknown node {identifier}")
    if len(name_matched) > 1:
        matching_ids = [node["node_id"] for node in name_matched]
        raise Exception(f"Ambigous name {identifier} is shared by nodes {matching_ids}")
    return name_matched[0]["node_id"]


def add(
    client: AppClient, workspace_id: str, node: str, duration: str, webhook_url: str, webhook_body: str
):
    node_id = _resolve_node_identifier(client, workspace_id, node)
    seconds = parse_duration(duration)
    if not seconds:
        print(f"invalid duration: {duration}")
        return
    try:
        json.loads(webhook_body)
    except:
        print(f"body must be valid json")
        return
    resp = client.post(
        "alert/add",
        json={
            "tenant_id": workspace_id,
            "node_id": node_id,
            "duration": seconds,
            "webhook_url": webhook_url,
            "webhook_body": webhook_body,
        },
    )
    resp.raise_for_status()
    print(json.dumps(resp.json()))


def delete(client: AppClient, workspace_id: str, alert_id: str):
    resp = client.post(
        "alert/delete",
        json={
            "tenant_id": workspace_id,
            "alert_id": alert_id,
        },
    )
    resp.raise_for_status()
    print(f"deleted alert {alert_id}")


def list(
    client: AppClient,
    workspace_id: str,
):
    resp = client.post(
        "alert/list",
        json={
            "tenant_id": workspace_id,
        },
    )
    resp.raise_for_status()
    alerts = resp.json()["alerts"]
    if len(alerts) == 0:
        print("no alerts configured")
        return
    print("Alert           Node            Trigger Url")
    for alert in alerts:
        alert_id = alert["id"]
        duration = alert["duration"]
        node = alert["node_id"]
        url = alert["webhook_url"]
        print(f"{alert_id}\t{node}\t{duration}s\t{url}")


def alert_subcommand(platform: PlatformEnvironment, argv):
    args = docopt(__doc__, argv=argv)
    try:
        workspace_id, user_key = load_current_workspace(platform)
        client = AppClient(platform=platform)
        client.workspace_login(user_key)
    except Exception as e:
        print(f"error: {e}")
        print(
            "Failed to load current workspace, please run 'tenzir-platform workspace select' first"
        )
        exit(1)

    try:
        if args["add"]:
            node = args["<node>"]
            duration = args["<duration>"]
            webhook_url = args["<webhook_url>"]
            webhook_body = args["<webhook_body>"]
            if webhook_body is None:
                webhook_body = f'{{"text": "Node $NODE_NAME disconnected for more than {duration}s"}}'
            assert json.loads(webhook_body), "body must be valid json"
            add(client, workspace_id, node, duration, webhook_url, webhook_body)
        elif args["delete"]:
            alert = args["<alert_id>"]
            delete(client, workspace_id, alert)
        elif args["list"]:
            list(client, workspace_id)
    except HTTPError as e:
        if e.response.status_code == 403:
            print(
                "Access denied. Please try re-authenticating by running 'tenzir-platform workspace select'"
            )
        else:
            print(f"Error communicating with the platform: {e}")



# SPDX-FileCopyrightText: (c) 2024 The Tenzir Contributors
# SPDX-License-Identifier: BSD-3-Clause

"""Tenzir Platform CLI.

Usage: tenzir-platform <command> [<args>...]
       tenzir-platform [--help] [--version]

Options:
  -h, --help                  Show this screen.
  --version                   Show version.

Commands:
   auth       Authenticate the current user
   workspace  Select the currently used workspace
   node       Interact with nodes
   alert      Configure alerts for disconnected nodes.
   admin      Administer local on-prem platform infrastructure.

See 'tenzir-platform <command> --help' for more information on a specific command.
"""

from docopt import docopt
from enum import Enum
import os
import requests
import requests.exceptions
import sys
import time
import importlib.metadata

from tenzir_platform.subcommand_auth import auth_subcommand
from tenzir_platform.subcommand_alert import alert_subcommand
from tenzir_platform.subcommand_workspace import workspace_subcommand
from tenzir_platform.subcommand_node import node_subcommand
from tenzir_platform.subcommand_admin import admin_subcommand
from tenzir_platform.helpers.environment import PlatformEnvironment

version = importlib.metadata.version("tenzir-platform")


def main():
    if len(sys.argv) == 1:
        sys.argv.append("--help")
    arguments = docopt(
        __doc__, version=f"Tenzir Platform CLI {version}", options_first=True
    )
    platform = PlatformEnvironment.load()
    argv = [arguments["<command>"]] + arguments["<args>"]
    if arguments["<command>"] == "auth":
        auth_subcommand(platform, argv)
    elif arguments["<command>"] == "alert":
        alert_subcommand(platform, argv)
    elif arguments["<command>"] == "workspace":
        workspace_subcommand(platform, argv)
    elif arguments["<command>"] == "node":
        node_subcommand(platform, argv)
    elif arguments["<command>"] == "admin":
        admin_subcommand(platform, argv)
    else:
        print("unknown subcommand, see 'tenzir-platform --help' for usage")


if __name__ == "__main__":
    main()



# SPDX-FileCopyrightText: (c) 2024 The Tenzir Contributors
# SPDX-License-Identifier: BSD-3-Clause

import hashlib
import os
import time

import jwt
import requests
from jwt import PyJWKClient
from typing import Optional, Any
from tenzir_platform.helpers.cache import filename_in_cache
from tenzir_platform.helpers.environment import PlatformEnvironment


class INVALID_API_KEY(Exception):
    pass


class ValidOidcToken:
    """The validated and decoded id_token.

    - Validate the presence of some claims required by the OIDC spec
    - Provide helper methods for checks"""

    def __init__(self, raw_oidc: dict[str, Any]) -> None:
        if "sub" not in raw_oidc or not isinstance(raw_oidc["sub"], str):
            raise INVALID_API_KEY("sub string required in OIDC token")
        self.user_id = raw_oidc["sub"]
        self._raw_oidc = raw_oidc

    def check_connection(self, connection: str) -> bool:
        return self.user_id.startswith(f"{connection}|")

    def get_claim_str(self, key: str, default: str) -> str:
        if key in self._raw_oidc and not isinstance(self._raw_oidc[key], str):
            raise INVALID_API_KEY(f"{key} is expected to be a string")
        return self._raw_oidc.get(key, default)

    @staticmethod
    def _is_list(val: Any) -> bool:
        return isinstance(val, list) and all(isinstance(x, str) for x in val)

    def get_claim_list(self, key: str, default: list[str]) -> list[str]:
        if key in self._raw_oidc and not ValidOidcToken._is_list(self._raw_oidc[key]):
            raise INVALID_API_KEY(f"{key} is expected to be a list of strings")
        return self._raw_oidc.get(key, default)

    def __str__(self) -> str:
        return self._raw_oidc.__str__()


class IdTokenClient:
    def __init__(self, platform: PlatformEnvironment):
        self.platform_environment = platform
        self.issuer = platform.issuer_url
        self.client_id = platform.client_id
        self.client_secret = platform.client_secret
        self.client_secret_file = platform.client_secret_file
        self.verbose = False
        discovery_url = f"{self.issuer.rstrip('/')}/.well-known/openid-configuration"
        discovered_configuration = requests.get(discovery_url).json()
        self.jwks_url = discovered_configuration["jwks_uri"]
        self.token_endpoint = discovered_configuration["token_endpoint"]
        self.device_authorization_endpoint = discovered_configuration[
            "device_authorization_endpoint"
        ]

    def validate_token(self, id_token: str) -> ValidOidcToken:
        """Verify the token using the audience specific to the CLI"""
        jwks_client = PyJWKClient(self.jwks_url)
        signing_key = jwks_client.get_signing_key_from_jwt(id_token)
        validated_token = jwt.decode(
            id_token,
            signing_key.key,
            algorithms=["RS256"],
            issuer=self.issuer,
            # for id tokens, the audience is the client_id
            audience=self.client_id,
        )
        return ValidOidcToken(validated_token)

    def reauthenticate_token(self, interactive: bool = True) -> str:
        if interactive:
            token_data = self._device_code_flow()
        else:
            token_data = self._client_credentials_flow()
        return self._unwrap_flow_result(token_data)

    def _device_code_flow(self) -> dict[str, str]:
        device_code_payload = {
            "client_id": self.client_id,
            # Request email by default since many auth rules are checked against the email address.
            "scope": "openid email",
        }

        device_code_response = requests.post(
            self.device_authorization_endpoint,
            data=device_code_payload,
        )

        if device_code_response.status_code != 200:
            raise Exception(
                f"Error generating the device code: {device_code_response.text}"
            )

        device_code_data = device_code_response.json()
        print(
            "1. On your computer or mobile device navigate to: ",
            device_code_data["verification_uri_complete"],
        )
        print(
            "2. Verify you're seeing the following code and confirm: ",
            device_code_data["user_code"],
        )
        print("3. Wait up to 10 seconds")

        token_payload = {
            "grant_type": "urn:ietf:params:oauth:grant-type:device_code",
            "device_code": device_code_data["device_code"],
            "client_id": self.client_id,
        }
        authenticated = False
        while not authenticated:
            token_response = requests.post(self.token_endpoint, data=token_payload)
            token_data = token_response.json()
            if token_response.status_code == 200:
                print("Authenticated!")
                break
            elif token_data["error"] not in ("authorization_pending", "slow_down"):
                print(token_data["error_description"])
                authenticated = True
            else:
                time.sleep(device_code_data["interval"])
        return token_data

    def _client_credentials_flow(self) -> dict[str, str]:
        if self.client_secret_file is not None:
            with open(self.client_secret_file, "r") as f:
                client_secret = f.read()
        elif self.client_secret is not None:
            client_secret = self.client_secret
        else:
            raise Exception(
                "Need a client secret in order to perform non-interactive login"
            )

        client_credentials_payload = {
            "grant_type": "client_credentials",
            "scope": "openid",
            "client_id": self.client_id,
            "client_secret": client_secret,
            "audience": self.client_id,
        }
        response = requests.post(self.token_endpoint, client_credentials_payload)
        response.raise_for_status()
        return response.json()

    def _unwrap_flow_result(self, token_data: dict[str, str]) -> str:
        id_token = token_data["id_token"]
        current_user = self.validate_token(id_token)
        if self.verbose:
            print(f"obtained id_token: {current_user}")
        self._store_id_token(id_token)
        return id_token

    def _filename_in_cache(self):
        return filename_in_cache(self.platform_environment, "id_token")

    def _store_id_token(self, token: str):
        filename = self._filename_in_cache()
        print(f"saving token to {filename}")
        os.makedirs(os.path.dirname(filename), exist_ok=True)
        with open(filename, "w") as f:
            f.write(token)

    def load_id_token(self, interactive: bool = True) -> str:
        filename = self._filename_in_cache()
        try:
            with open(filename, "r") as f:
                token = f.read()
            self.validate_token(token)
            return token
        except Exception:
            print("could not load valid token from cache, reauthenticating")
        return self.reauthenticate_token(interactive=interactive)



# SPDX-FileCopyrightText: (c) 2024 The Tenzir Contributors
# SPDX-License-Identifier: BSD-3-Clause

from pydantic_settings import BaseSettings
from typing import Optional

API_ENDPOINT = "https://rest.tenzir.app/production-v1"
ISSUER_URL = "https://tenzir.eu.auth0.com/"
CLIENT_ID = "vzRh8grIVu1bwutvZbbpBDCOvSzN8AXh"


# All values can be set using environment variables like
#
#   `TENZIR_PLATFORM_CLI_API_ENDPOINT=https://tenzir.example`
#
# (and also can only set this way, because we don't provide
# separate command-line options)
class PlatformEnvironment(BaseSettings):
    # The remote API endpoint of the platform.
    api_endpoint: str = API_ENDPOINT

    # An arbitrary short string to identify this environment
    # in the local cache directory.
    stage_identifier: str = "prod"

    # Configuration of the OIDC for authentication.
    issuer_url: str = ISSUER_URL
    client_id: str = CLIENT_ID

    # A client secret is only necessary for non-interactive logins
    # using the client credentials flow.
    client_secret: Optional[str] = None
    client_secret_file: Optional[str] = None

    # Enable more verbose print statements.
    verbose: bool = False

    @staticmethod
    def load():
        return PlatformEnvironment(
            _env_prefix="TENZIR_PLATFORM_CLI_", _env_nested_delimiter="__"
        )
